__version__ = 4, 2, 13
__author__ = '竹永康'
__author_email__ = 'zhuyk4@lenovo.com'


def __init__(config: dict):
    for icls, iconf in config.items():
        if icls[0] != 'i':
            continue
        try:
            icls = getattr(sys.modules[__name__], icls)
            __import__(icls.__doc__)
            imodule = sys.modules[icls.__doc__]
        except (AttributeError, ModuleNotFoundError):
            continue

        general_params: dict = iconf.pop('init', {})

        for iname, params in iconf.items():
            for item in general_params.items():
                params.setdefault(*item)

            iobj = imodule.__init__(iname, **params)
            setattr(icls, iname, iobj)

    setattr(sys.modules[__name__], 'inf', config)


class ialiyun:
    """isddc_aliyun"""

    def __new__(cls, iname: str = None, **params) -> 'Aliyun':
        """Get a Aliyun object.

        @param iname:
            Assigns the initialized Aliyun object to a
            variable in the current module, if not None.

        @param params:
            Initialization parameters of Aliyun.
            Sample1:
                name='SendSms',
                ak=None,
                secret=None,
                region_id='cn-hangzhou',
                domain='dysmsapi.aliyuncs.com',
                version='2017-05-25',
                SignName='歌琪怡玲萍云',
                TemplateCode='SMS_199195663',
                PhoneNumbers='1388888888',
                timeout=60,
                method='post',
                protocol_type='https',
                accept_format='json'
            Sample2:
                name='SingleSendMail',
                ak=None,
                secret=None,
                region_id='cn-hangzhou',
                domain='dm.aliyuncs.com',
                version='2015-11-23',
                AddressType=1,
                ReplyToAddress=True,
                AccountName='trigger@mail.gqylpy.com',
                ToAddress='zhuyk4@lenovo.com',
                FromAlias='gqylpy-base',
                timeout=60,
                method='post',
                protocol_type='https',
                accept_format='json'

        @return: Aliyun(@param(params))
        """

    class Aliyun:
        def __init__(self, action: dict):
            pass

    @staticmethod
    def send_sms(**active_params):
        pass

    @staticmethod
    def send_mail(**active_params):
        pass


class iapi:
    """isddc_api"""

    def __new__(cls, iname: str = None, **params) -> 'IsddcAPI':
        """Get a IsddcAPI object.

        @param iname:
            Assigns the initialized logging.Logger object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of IsddcAPI, sample:
                api='10.121.134.109:30870',
                keystone='10.121.134.109:31500',
                username=None,
                password=None,
                tenant_id=None,
                http_timeout=60,
                token_timeout=60*59

        @return: IsddcAPI(**@param(params))
        """

    class IsddcAPI:
        """Born for OI isddc-api."""

        def __init__(
                self,
                *,
                api: str,
                keystone: str,
                username: str,
                password: str,
                tenant_id: str = None,
                tenant_name: str = None,
                http_timeout: int = 60,
                token_timeout: int = 60 * 59,
        ):
            """
            @param api:           address:port of isddc-api service.
            @param keystone:      address:port of keystone service.
            @param username:      Keystone auth username.
            @param password:      Keystone auth password.
            @param tenant_id:     Keystone auth tenant id, or @param(tenant_name), high priority.
            @param tenant_name:   Keystone auth tenant name, or @param(tenant_id).
            @param http_timeout:  Http timeout.
            @param token_timeout: Token timeout.
            """

        def get(
                self,
                uri: str,
                *,
                params: dict = None,
                **kw
        ) -> 'requests.Response':
            """"Get data from OI isddc-api."""

        def post(
                self,
                uri: str,
                *,
                data: dict = None,
                json: dict = None,
                **kw
        ) -> 'requests.Response':
            """Post data to OI isddc-api."""

        def report(
                self,
                data: dict,
                *,
                dtype: 'enum(metrics, inventories)'
        ) -> 'requests.Response':
            """Report data, metrics or inventories."""

        @staticmethod
        def check_status(res: 'requests.Response') -> 'requests.Response':
            """If the response status code is greater
            than 400, an HttpError is thrown."""

    @staticmethod
    def get(
            uri: str,
            *,
            params: dict = None,
            iname: IsddcAPI = None,
            **kw
    ) -> 'requests.Response':
        """Get data from OI isddc-api."""

    @staticmethod
    def post(
            uri: str,
            *,
            data: dict = None,
            json: dict = None,
            iname: IsddcAPI = None,
            **kw
    ) -> 'requests.Response':
        """Post data to OI isddc-api."""

    @staticmethod
    def report(
            data: dict,
            *,
            dtype: 'enum(metrics, inventories)',
            iname: IsddcAPI = None
    ) -> 'requests.Response':
        """Report data, metrics or inventories."""


class iasync:
    """isddc_multitask.IsddcAsyncTasks"""

    def __new__(cls, *a, **kw):
        raise NotImplementedError

    @classmethod
    def at(
            cls,
            async_func: 'Callable',
            *a,
            namespace: str = None,
            **kw
    ):
        """Submit an asynchronous function and
        store it in the specified namespace."""

    @classmethod
    def batch_at(
            cls,
            async_func: 'Callable',
            *params
    ) -> str:
        """Batch submit async task."""

    @classmethod
    def run_until_complete(
            cls,
            namespace: str = None,
            *,
            timeout: int = None,
            return_when: str = 'ALL_COMPLETED'
    ) -> tuple:
        """Run all asynchronous function in the
        specified namespace until completed."""

    @classmethod
    def get(
            cls,
            namespace: str = None,
            *,
            timeout: int = None,
            return_when: str = 'ALL_COMPLETED'
    ) -> 'Iterator':
        """Run all asynchronous function in the specified
        namespace until complete, and get all task results."""

    @staticmethod
    def random_namespace() -> str:
        """Return a random uuid4 string."""

    @staticmethod
    def run_async_func(async_func: 'Callable', *a, **kw) -> 'Any':
        """Run an asynchronous function and return its result."""

    @staticmethod
    def asyncfunc_to_func(async_func):
        """The decorated asynchronous function
        will be executed synchronously."""


class ibase:
    """isddc_baseclass"""

    class SingletonMode:
        """Inherit me, you will
        become a singleton class."""

    class DictMode:
        """Inherit me, You will have
        the grammar of the 'dict'."""

        def __getitem__(self, name):
            pass

        def __setitem__(self, name, value):
            pass

        def __delitem__(self, name):
            pass

    class IterMode:
        """Inherit me, your object can be iterated,
        and can be called by the 'if in' statement."""

        def __iter__(self):
            pass

    class ObjectMode(type):
        """To essentially turn a class into an object."""

    class FuncMode(type):
        """Enable classes to be called as functions.
        Instantiating class no longer get the object, it
        directly gets the return value of its __call__ method."""


class ibootparams:
    """isddc_bootparams"""

    @staticmethod
    def filter(*params, **default_params) -> dict:
        """Get boot parameters.

        @param params:         Gets which boot parameters.
        @param default_params: Boot parameters with default values.

        Boot parameters with default values are available through **default_params.
        """

    @staticmethod
    def get(param: str, define=None) -> str:
        """
        Get a boot parameter, or return @param(define) if none exists

        :param param:  Boot parameter name.
        :param define: Default value.
        :return:       The boot parameter or @param(define)
        """


class icmd:
    """isddc_command.IsddcCommand"""

    """
    Executes the incoming command directly at instantiation time, and saves:
        self.code       = Command exit code
        self.raw_output = Command output
    """
    code: int
    raw_output: str

    def __init__(self, cmd: str, *, timeout: int = None):
        """
        @param cmd:     Command string.
        @param timeout: Command timeout, Default never times out.
        """

    def raise_if_error(self):
        if self.code != 0:
            raise self.CommandError

    @property
    def status(self) -> bool:
        if self.code == 0:
            return True
        return False

    @property
    def output(self) -> str:
        return process(self.raw_output)

    @property
    def status_output(self) -> tuple:
        return self.status, self.output

    @property
    def code_output(self) -> tuple:
        return self.code, self.output

    @property
    def output_else_raise(self) -> str:
        if self.code == 0:
            return self.output
        raise self.CommandError

    @property
    def output_else_define(self, define: 'Any' = None):
        if self.code == 0:
            return self.output
        return define

    def is_contain_string(self, string: str) -> bool:
        if string in self.raw_output:
            return True
        return False

    def output_if_contain_string_else_raise(self, string: str) -> str:
        if string in self.raw_output:
            return self.output
        raise self.CommandError

    def table_output_to_dict(self, **kw) -> list:
        """Can turn command results printed in tabular format
        into dictionaries, such as `kubectl get pod -n isddc`."""

    @classmethod
    def table2dict(cls, table: str, *, split: str = None) -> list:
        """Can turn command results printed in tabular format
        into dictionaries, such as `kubectl get pod -n isddc`."""

    class CommandError(Exception):
        pass


class icrypto:
    """isddc_crypto"""

    class IsddcCrypto:
        """A reversible encryption technique based on Unicode Code."""

        def __init__(self, offset: int = -1):
            """
            @param offset:
                Specifies the Unicode Code offset.
                Encryption and decryption must use the same value.
            """

        def encrypt(self, cleartext: str) -> str:
            """
            @param cleartext: The string to be encrypt.
            @return:          The encrypted value.
            """

        def decrypt(self, ciphertext: str) -> str:
            """
            @param ciphertext: The string to be decrypt.
            @return:           The decrypted value.
            """

    class MD5:

        def __new__(cls, data: 'Any' = None):
            """
            @param data: Data to be encrypt.

            if data is None:
                return object.__new__(cls)
            else:
                return md5-value of data.
            """

        def update(self, data: 'Any'):
            """Add or append a segment of data to encrypt."""

        def update_str(self, data: str):
            """Add or append a string to encrypt."""

        @property
        def hexdigest(self) -> str:
            """Execute encryption program and return md5-value."""

    def encrypt(self, cleartext: str, *, offset: int = -1) -> str:
        """Encrypts the string and returns the encrypted value."""
        return self.IsddcCrypto(offset).encrypt(cleartext)

    def decrypt(self, ciphertext: str, *, offset: int = -1) -> str:
        """Decrypt the ciphertext and return the decrypted value."""
        return self.IsddcCrypto(offset).decrypt(ciphertext)


class idbpool:
    """isddc_dbpool"""

    def __new__(cls, iname: str = None, **params) -> 'dbutils.pooled_db.PooledDB':
        """Get a PooledDB object.

        @param iname:
            Assigns the initialized PooledDB object to
            a variable in the current module, if not None.

        @param params:
            Initialization parameters of PooledDB, sample:
                creator='pymysql',
                db='isddc',
                host='localhost',
                port=3306,
                user='isddc',
                passwd='user@isddc',
                charset='utf8',
                autocommit=False,
                use_unicode=None,
                read_timeout=None,
                write_timeout=None,
                connect_timeout=None,
                mincached=0,
                maxcached=1,
                maxshared=0,
                maxconnections=2,
                blocking=True,
                maxusage=None,
                setsession=None,
                ping=1
            Parameters describe:
                @param.creator
                    数据库模块名称
                @param.**kwargs
                    数据库膜拜参数
                @param.mincached
                    初始化时，连接池中至少创建的空闲的连接，0表示不创建
                @param.maxcached
                    连接池中最多闲置的连接，0和None表示不限制
                @param.maxshared
                    连接池中最多共享的连接数量，0和None表示全部共享（这个参数不管指定多少，最后都会是0）
                    因为pymysql和MySQLdb等模块的threadsafety都为1，
                    所以值无论设置为多少，_maxshared永远为0，所以永远是所有连接都共享
                @param.maxconnections
                    连接池允许的最大连接数，0和None表示不限制连接数
                @param.blocking
                    连接池中如果没有可用连接后，是否阻塞等待，True:等待 False:不等待，直接报错
                @param.maxusage
                    一个连接最多可被使用的次数，None表示无限制
                @param.setsession
                    开始会话前执行的命令列表，如: ['set dtype to ...', 'set time zone ...']
                @param.ping
                    Ping MySQL服务端，检查服务是否可用
                    0: 从来没有  1: 从池中提取时（默认） 2: 创建游标时  4: 执行查询时  7: 总是

        @return: PooledDB(**@param(params))
        """

    class mysql:

        def __new__(
                cls,
                sql: str = None,
                *,
                one: bool = False,
                dtype: 'Union[list, dict]' = dict,
                iname: 'dbutils.pooled_db.PooledDB' = None
        ):
            """
            if sql is not None:
                Execute the sql and return the results.
            else:
                Continue calling __init__ and get an instance of mysql.
            """
            pass

        def __init__(
                self,
                *,
                dtype: 'Union[list, dict]' = dict,
                iname: 'dbutils.pooled_db.PooledDB' = None
        ):
            pass

        def __call__(self, s: str) -> int:
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def __invert__(self):
            pass

        @staticmethod
        def upgrade_dtype(data: list = None) -> list:
            pass

        @property
        def fetchone(self) -> 'Union[tuple, dict]':
            pass

        @property
        def fetchall(self) -> 'Union[tuple, dict]':
            pass

        @property
        def last_insert_id(self) -> int:
            """Returns the id of the data that
            was last inserted by `self.conn`."""


class idec:
    """isddc_decorator"""

    @staticmethod
    def TryExcept(
            etype: 'Union[type, tuple]',
            *,
            name: str = None,
            ignore: bool = False,
            exc_return: 'Any' = None,
            exc_callback: 'Callable' = None,
            output_original_exc: bool = False
    ):
        """Exception handler.
        The function that is decorated will have exception handling capabilities.

        @param etype:               Which exceptions to handle.
        @param name:                Exception identification, which is displayed in the exception output.
        @param ignore:              Whether to ignore an exception or not.
                                    If so, no exception information will be logged.
        @param exc_return:          The value returned after the exception was raised.
        @param exc_callback:        The function that is called after the exception is raised.
        @param output_original_exc: Whether to print the original exception message.

    #-- Simple demo:
        @idec.TryExcept(ValueError)
        def func():
            int('a')
    """

    @staticmethod
    def Retry(
            name: str = None,
            *,
            count: int = None,
            cycle: int = 0,
            output_original_exc: bool = False,
            retry_exc: 'Union[type, tuple]' = Exception
    ):
        """
        When an exception is raised, attempt to reexecute.

        @param name:                Exception identification which is displayed in the exception output.
        @param count:               Total number of executions, default N.
        @param cycle:               Seconds between each attempt, default 0.
        @param output_original_exc: Whether to print the original exception message.
        @param retry_exc:           Which exceptions are supported to retry.

        #-- Simple demo:
            @idec.Retry(count=3, cycle=1)
            def func():
                int('a')

        #-- Use in conjunction with TryExcept will support
            both exception retry and exception handling:
            @idec.TryExcept(ValueError)
            @idec.Retry(count=3, cycle=1)
            def func():
                int('a')
        """

    @staticmethod
    def WhileTrue(
            cycle: 'Union[int, float]' = 0,
            *,
            minus_run_time: bool = False,
            wait_first: bool = False
    ):
        """
        The decorated function will be executed in an infinite
        loop, Can end the loop by returning a string '_break_'.

        @param cycle:          Loop interval.
        @param minus_run_time: Loop interval minus function run time.
                               If True, the cycle period will be more accurate.
        @param wait_first:     If True, the first execution will wait for @param(cycle) seconds.

        #-- Simple demo:
            @idec.WhileTrue(cycle=1, minus_run_time=True)
            def func():
                print(itime.Now.str)

        #-- To exit the loop, use syntax `return '_break_'`:
            @idec.WhileTrue(cycle=1, minus_run_time=True)
            def func():
                print(itime.Now.str)
                return '_break_'

        #-- Use in conjunction with TryExcept will exceptions
            that are raised in each cycle can be handled:
            @idec.WhileTrue(cycle=1, minus_run_time=True)
            @idec.TryExcept(Exception)
            def func():
                print(itime.Now.str)
                int('a')
        """

    @staticmethod
    def RunTime(name: str = None):
        """
        Record the execution time of the function.

        @param name: Function identification, reflected in the output

        #-- Simple demo:
            @idec.RunTime()
            def func():
                time.sleep(1.234)
        """

    @staticmethod
    def TestFuncSpeed(count: int, *, keep: int = 7):
        """Test the execution speed of the function."""

    @staticmethod
    def BeforeFunc(before_func: 'Callable', *, enable: bool = True):
        """
        Execute another function before another function.

        @param before_func: A func.
        @param enable:      If False, then @param(before_func) is not called.

        #-- Simple demo:
            def before_func():
                print('exec before_func')

            @idec.idec.BeforeFunc(before_func)
            def func():
                print('exec func')

        #-- When have parameters:
            def before_func(param1, param2):
                print('exec before_func', param1, param2)

            @idec.BeforeFunc(before_func)
            def func(param1, param2):
                print('exec func', param1, param2)

        #-- Modify parameters in 'before_func':
            def before_func(param1, param2):
                print('exec before_func', param1, param2)
                return {'param1': param1.upper()}

            @idec.BeforeFunc(before_func)
            def func(param1, param2):
                print('exec func', param1, param2)
        """

    @staticmethod
    def AfterFunc(
            after_func: 'Callable',
            *,
            use_raw_params: bool = False,
            enable: bool = True
    ):
        """
        Executes another function after another function.

        @param after_func:     A func.
        @param use_raw_params: See the above usage.
        @param enable:         If False, then @param(after_func) is not called.

        #-- Simple demo:
            def after_func():
                print('exec after_func')

            @idec.AfterFunc(after_func)
            def func():
                print('exec func')

        #-- When have parameters:
            def after_func(param1, param2):
                print('exec after_func', param1, param2)

            @idec.AfterFunc(after_func)
            def func(param1, param2):
                print('exec func', param1, param2)
                return param1.upper(), param2.upper()

        #-- @param(use_raw_params) usage:
            def after_func(param1, param2):
                print('exec after_func', param1, param2)

            @idec.AfterFunc(after_func, use_raw_params=True)
            def func(param1, param2):
                print('exec func', param1, param2)
        """

    @staticmethod
    def TryExceptAsync(
            etype: 'Union[type, tuple]',
            *,
            name: str = None,
            ignore: bool = False,
            exc_return: 'Any' = None,
            exc_callback: 'Callable' = None,
            output_original_exc: bool = False
    ):
        """Exception handler.
        The function that is decorated will have exception handling capabilities.

        @param etype:               Which exceptions to handle.
        @param name:                Exception identification, which is displayed in the exception output.
        @param ignore:              Whether to ignore an exception or not.
                                    If so, no exception information will be logged.
        @param exc_return:          The value returned after the exception was raised.
        @param exc_callback:        The function that is called after the exception is raised.
        @param output_original_exc: Whether to print the original exception message.

        #-- Simple demo:
            @idec.TryExceptAsync(ValueError)
            async def func():
                int('a')
        """

    @staticmethod
    def RetryAsync(
            name: str = None,
            *,
            count: int = None,
            cycle: int = None,
            output_original_exc: bool = False,
            retry_exc: 'Union[type, tuple]' = Exception
    ):
        """
        When an exception is raised, attempt to reexecute.

        @param name:                Exception identification which is displayed in the exception output.
        @param count:               Total number of executions, default N.
        @param cycle:               Seconds between each attempt, default 0.
        @param output_original_exc: Whether to print the original exception message.
        @param retry_exc:           Which exceptions are supported to retry.

        #-- Simple demo:
            @idec.RetryAsync(count=3, cycle=1)
            async def func():
                int('a')

        #-- Use in conjunction with TryExcept will support
            both exception retry and exception handling:
            @idec.TryExceptAsync(ValueError)
            @idec.RetryAsync(count=3, cycle=1)
            async def func():
                int('a')
        """


class idict(dict):
    """isddc_dict.IsddcDict"""

    """idict == dict
    I inherited 'dict', my main function is to
    allow the dict to get or set values by `d.key`.
    """

    def __init__(self, *a, **kw):
        pass

    def __getattribute__(self, name):
        pass

    def __setattr__(self, name, value):
        pass

    def __delattr__(self, name):
        pass


class ifs:
    """isddc_filesystem"""

    class FileSystem:
        """FileSystem was FileDataOperator in a previous life.

        db = FileSystem('db/')

        # Write content to file 'db/test.txt'
        db['test.txt'].content = ...

        # Read content from file 'db/test.txt'
        text = db['test.txt'].content

        # Delete file 'db/test.txt'
        del db['test.txt']

        # Deep directory:
        db['dir1/dir2/test.txt'] = ...
        text = db['dir1/dir2/test.txt'].content

        # Get file information:
        info = db['test.txt'].{name,atime,mtime,ctime,size}
        """

        def __init__(self, root: str):
            """
            @param root:
                Specify a directory path, it could be a relative path.
                If the specified directory does not exist, it is created.
            """

        def __getattr__(self, rel_path: str):
            pass

        def __setattr__(self, rel_path: str, data):
            pass

        def __delattr__(self, rel_path: str):
            pass

        def __getitem__(self, rel_path: str):
            pass

        def __setitem__(self, rel_path: str, data):
            pass

        def __delitem__(self, rel_path: str):
            pass

        def __iter__(self):
            pass

    class FileDataOperator:
        """
        db = ifs.FileDataOperator('db/')

        # Write content to file 'db/test.txt'
        db['test.txt'] = ...

        # Read content from file 'db/test.txt'
        text = db['test.txt']

        # Delete file 'db/test.txt'
        del db['test.txt']

        # Deep directory:
        db['dir1/dir2/test.txt'] = ...
        text = db['dir1/dir2/test.txt']
        """

        def __init__(self, root: str):
            """
            @param root:
                Specify a directory path, it could be a relative path.
                If the specified directory does not exist, it is created.
            """
            self.root: str = ifs.gendir(root)
            self.path: dict = ifs.fetch_deep_path(self.root)

        def __getattr__(self, name):
            pass

        def __getitem__(self, name):
            pass

        def __setitem__(self, name, value):
            pass

        def __delitem__(self, name):
            pass

        def __iter__(self):
            pass

    @staticmethod
    def Filetor(
            file: str,
            data: 'Any' = None,
            *,
            ftype: 'enum(json, yaml, obj, txt)' = None,
            encoding: str = None,
            **kw
    ):
        """
        Writes data to or reads data from a file based on its type.

        @param file:     A file path.
        @param data:     The data to be written to the file, if None,
                         the data is read from the file and return.
        @param ftype:    Specifies File type, otherwise is determined by file suffix name.
        @param encoding: Open the encoding of the file, default UTF-8

        if @param(data) is None:
            Read contents of @param(file) according to @param(ftype) and return.
        else:
            Write @param(data) to @param(file) according to @param(ftype).
        """

    @staticmethod
    def fetch_deep_path(path: str = None) -> dict:
        """Get all file and directory paths under @param(path)."""

    @staticmethod
    def joinpath2abs(*a) -> str:
        """Return its absolute path after concatenating the path."""

    @staticmethod
    def dirname(path: str, *, level: int = 1) -> str:
        """Culling directories from the
        end according to `level`, usually
        used to get the project root path.
        """

    @staticmethod
    def gendir(*a) -> str:
        """Generate path
        Generates a directory based on the passed in value,
        creates it if it does not exist, and returns it.
        """

    @staticmethod
    def isabs(path: str) -> bool:
        """Whether @param(path) is an absolute path."""

    @staticmethod
    def isdir(path: str) -> bool:
        """Whether @param(path) is a directory."""

    @staticmethod
    def isfile(path: str) -> bool:
        """Whether @param(path) is a file."""

    @staticmethod
    def exists(path: str) -> bool:
        """Whether the @param(path) exists."""

    @staticmethod
    def joinpath(*a) -> str:
        """Splice @param(*a) into a path."""

    @staticmethod
    def abspath(path: str) -> str:
        """Get the full path of @param(path)."""

    @staticmethod
    def basename(path: str) -> str:
        """Gets the filename of @param(path), and
        returns null if the path is not a file."""


class itools:
    """isddc_tools"""

    linux: bool = None
    container: bool = None

    @staticmethod
    def DictInterProcess(data: dict, func: 'Callable'):
        """Dictionary internal processing
        Used to batch process values inside a dictionary.
        @param(func) process @param(data).

        #-- Simple demo:
            data = {'a': 'a', 'b': 'b'}

            def func(name, value):
                if name == 'b':
                    return value.upper()

            DictInterProcess(data, func)
            print(data)
        """

    @staticmethod
    def underline2hump(name: str) -> str:
        """Change the underscore name to the hump name."""

    @staticmethod
    def hump2underline(name: str) -> str:
        """Change the hump name to the underscore name."""

    @staticmethod
    def get_caller(level: int = 1) -> str:
        """Gets the location where the current function was called."""

    @staticmethod
    def prt(
            *args,
            color: int = 31,
            font: int = 0,
            end: str = '\n',
            sep: str = ' ',
            file: open = None
    ):
        """
        Custom print font and color.

        Color:
            31-red 32-green 34-blue 37-Gray 38-black
        Font:
            0-normal 1-bold
        sep:
            string inserted between values, default a space.
        """

    @staticmethod
    def avg(numbers: list) -> float:
        """Return average value of @param(numbers)."""

    @staticmethod
    def count_occurrences(data: list, value: 'Any') -> int:
        """Return number of occurrences of @param(value) in @param(data)."""

    @staticmethod
    def max_value_index(values: list) -> int:
        """Return the index of the maximum value in @param(values)."""


class importer:
    """isddc_import"""

    @staticmethod
    def module(module: str):
        """
        @param module: A module path of string, e.g. <package>.<module>
        @return:       A object of @param(module).
        """

    @classmethod
    def name(cls, name: str):
        """
        @param name: A variable path of string, e.g. <package>.<module>.<variable>
        @return:     A variable.
        """

    @classmethod
    def object(cls, class_path: str, *a, **kw):
        """
        @param class_path: A class path of string, e.g. <package>.<module>.<class>
        @param a:          Class instantiates parameters.
        @param kw:         Class instantiates parameters.
        @return:           A object of @param(class_path).
        """

    @classmethod
    def trying(cls, module: str, default: 'Any' = None):
        """
        @param module:  A module path of string, e.g. <package>.<module>
        @param default: This value is returned if it results in an error.
        @return:        A object of @param(module), or @param(default)
        """


class influx:
    """isddc_influx"""

    def __new__(cls, iname: str = None, **params) -> 'influxdb.InfluxDBClient':
        """Get a InfluxDBClient object.

        @param iname:
            Assigns the initialized InfluxDBClient object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of InfluxDBClient, sample:
                host='localhost',
                port=8086,
                database='isddc'

        @return: InfluxDBClient(**@param(params))
        """

    @staticmethod
    def query(
            sql: str,
            *,
            one: bool = False,
            epoch: str = None,
            dtype: 'Union[dict, list]' = dict,
            iname: 'influxdb.InfluxDBClient' = None,
            **kw
    ) -> 'Union[dict, list]':
        """
        Encapsulate the influxdb.InfluxDBClient.query method.

        @param sql:    SQL statement string.
        @param one:    Returns a piece of data.
        @param epoch:  The time format.
        @param dtype:  The data type returned, dict or str
        @param client: Specified database, the first is used by default.
        """

    @staticmethod
    def write_points(
            data: list,
            *,
            timep: str = None,
            iname: 'influxdb.InfluxDBClient' = None,
            **kw
    ) -> bool:
        """Encapsulate the influxdb.InfluxDBClient.write_points method."""


class ikafka:
    """isddc_kafka"""

    def __new__(
            cls, iname: str = None, **params
    ) -> 'Union[pykafka.Topic, pykafka.KafkaClient]':
        """Get a pykafka.Topic object.

        @param iname:
            Assigns the initialized pykafka.Topic object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of pykafka.Topic, sample:
                hosts='localhost:31092',
                topic='isddc'

        @return: pykafka.Topic(**@param(params))
        """

    @staticmethod
    def send(msg: 'Any', *, iname: 'pykafka.Topic' = None):
        """Simple Kafka message sender."""


class ikube:
    """isddc_kube"""

    def __new__(cls, iname: str = None, **params) -> 'kubernetes.client.CoreV1Api':
        """Get a kube_client.CoreV1Api object.

        @param iname:
            Assigns the initialized kube_client.CoreV1Api object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of kube_client.CoreV1Api, sample:
                api_server='localhost:6443'
                token=None,
                protocol='https',
                verify_ssl=False,
                load_api='CoreV1Api, AppsV1Api'

        @return: kubernetes.client.CoreV1Api
        """


class ilog:
    """isddc_log"""

    def __new__(cls, iname: str = None, **params) -> 'logging.Logger':
        """Get a logging.Logger object.

        @param iname:
            Assigns the initialized logging.Logger object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of logging.Logger, sample:
                level='NOTSET',
                output='stream,file',
                handler='/var/log/isddc.log',
                datefmt='%F %T',
                logfmt='[%(asctime)s] [%(levelname)s] %(message)s'

        @return: logging.Logger(**@param(params))
        """

    @staticmethod
    def debug(msg: str, *, iname: 'logging.Logger' = None):
        pass

    @staticmethod
    def info(msg: str, *, iname: 'logging.Logger' = None):
        pass

    @staticmethod
    def warning(msg: str, *, iname: 'logging.Logger' = None):
        pass

    @staticmethod
    def error(msg: str, *, iname: 'logging.Logger' = None):
        pass

    @staticmethod
    def critical(msg: str, *, iname: 'logging.Logger' = None):
        pass


class imongo:
    """isddc_mongo"""

    def __new__(
            cls, iname: str = None, **params
    ) -> 'Union[pymongo.MongoClient, pymongo.database.Database]':
        """Get a MongoClient object.

        @param iname:
            Assigns the initialized MongoClient object to
            a variable in the current module, if not None.

        @param params:
            Initialization parameters of MongoClient, sample:
                host='localhost',
                port=27017,
                db='isddc'

        @return: MongoClient(**@param(params)) or Database
        """

    @staticmethod
    def get_collection(
            collection: str,
            *,
            iname: 'Union[pymongo.MongoClient, pymongo.database.Database]' = None
    ) -> 'pymongo.collection.Collection':
        """Get a collection and return the collection object."""


class imt:
    """isddc_multitask"""

    class IsddcProcessPool:

        def __init__(
                self,
                name: str = None,
                *,
                processes: int = None,
                **kw
        ):
            """
            Create a Process pool.
            @param name:      Pool name.
            @param processes: Process number.
            """

        def at(
                self,
                func: 'Callable',
                *a,
                namespace: str = None,
                callback: 'Callable' = None,
                error_callback: 'Callable' = None,
                **kw
        ) -> 'multiprocessing.pool.AsyncResult':
            """
            Submit a task to the pool.
            @param func:           A function.
            @param a:              Position parameters of @param(func).
            @param namespace:      Specifies the  name space of @param(func)
                                   and creates it if it does not exist.
            @param callback:       Callback function of @param(func).
            @param error_callback: Error callback function of @param(func).
            @param kw:             Keyword parameters of @param(func).
            @return:               Result object.
            """

        def batch_at(
                self,
                func: 'Callable',
                *params,
                callback: 'Callable' = None,
                error_callback: 'Callable' = None
        ) -> str:
            """Batch submit tasks to the pool, and returns
            the namespace in which these tasks reside."""

        def block_until_complete(
                self,
                namespace: str = None,
                *,
                delay_time: int = 0,
                timeout: int = None,
                ignore_timeout: bool = False,
                ignore_namespace: bool = False,
                cycle: 'Union[int, float]' = None
        ) -> list:
            """
            Wait for all tasks in the specified namespace to complete.
            @param namespace:        A namespace.
            @param delay_time:       How many seconds to wait before entering the loop check phase.
            @param timeout:          Tasks timeouts.
            @param ignore_timeout:   If true, a task timeout will throw an exception.
            @param ignore_namespace: Returns an empty list if the namespace does not exist, if it is True.
            @param cycle:            The loop checks the duration of the task, default is 0.1 seconds.
            @return:                 Tasks result.
            """

        @staticmethod
        def get(
                result: 'multiprocessing.pool.AsyncResult',
                *,
                delay_time: int = None,
                timeout: int = None,
                ignore_timeout: bool = False,
                cycle: int or float = None
        ) -> 'Any':
            """
            Get the result of an asynchronous task.
            @param result:           Asynchronous task object.
            @param delay_time:       How many seconds to wait before entering the loop check phase, default 0 seconds.
            @param timeout:          Tasks timeouts.
            @param ignore_timeout:   If true, a task timeout will throw an exception.
            @param cycle:            The loop checks the duration of the task, default 0.1 seconds.
            @return:                 The @param(result) value.
            """

        def batch_at_and_wait_complete(
                self,
                func: 'Callable',
                *params,
                **kw
        ) -> list:
            pass

        @staticmethod
        def block(sleep: int):
            """The actual execution will be time.sleep()."""
            pass

        @staticmethod
        def random_namespace() -> str:
            """Return a random uuid4 string."""

    class IsddcThreadPool(IsddcProcessPool):
        def __init__(
                self,
                name: str = None,
                *,
                processes: int = None,
                **kw
        ):
            """
            Create a thread pool.
            @param name:      Pool name.
            @param processes: Thread number.
            """


class imysql:
    """isddc_mysql"""

    def __new__(cls, iname: str = None, **params) -> 'pymysql.Connection':
        """Get a pymysql.Connection object.

        @param iname:
            Assigns the initialized pymysql.Connection object
            to a variable in the current module, if not None.

        @param params:
            Initialization parameters of pymysql.Connection, sample:
                db='isddc',
                host='localhost',
                port=3306,
                user='isddc',
                passwd='user@isddc',
                charset='utf8',
                autocommit=False,
                use_unicode=None,
                read_timeout=None,
                write_timeout=None,
                connect_timeout=None

        @return: pymysql.Connection(**@param(params))
        """

    @staticmethod
    def sql(
            s: str,
            *,
            one: bool = False,
            dtype: 'Union[dict, tuple]' = dict,
            iname: 'pymysql.Connection' = None
    ):
        """Execute an SQL statement.
        @param s:     SQL statement.
        @param one:   If true, the first data in the query set will be
                      returned, which is suitable for single data query.
        @param dtype: Data type, dictionary or tuple.
        @param conn:  Database connection, First initialized connection is used by default.
        @return:      Queryset or commit result.
        """

    class Transaction:
        """Execute MySQL transactions."""

        def __init__(
                self,
                dtype: 'Union[dict, tuple]' = dict,
                *,
                iname: 'pymysql.Connection' = None
        ):
            """
            @param dtype: Data type, dictionary or tuple.
            @param iname:  Database connection, First initialized connection is used by default.
            """

        def __call__(self, s: str) -> int:
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def __invert__(self):
            pass

        @staticmethod
        def upgrade_dtype(data: list = None) -> list:
            pass

        @property
        def fetchone(self) -> 'Union[tuple, dict]':
            pass

        @property
        def fetchall(self) -> 'Union[tuple, dict]':
            pass

        @property
        def last_insert_id(self) -> int:
            """Returns the id of the data that
            was last inserted by `self.conn`."""


class ioverfunc:
    """isddc_overfunc"""

    @staticmethod
    def at(func: 'Callable', *a, **kw):
        """
        Add a function that will execute after the main program is finished.
        @param func: A function to be executed at after the main program is finished.
        @param a:    Function parameters.
        @param kw:   Function parameters.
        """


class issh:
    """isddc_ssh"""

    def __new__(cls, iname: str = None, **params) -> 'paramiko.SSHClient':
        """Get a SSHClient object.

        @param iname:
            Assigns the initialized SSHClient object to
            a variable in the current module, if not None.

        @param params:
            Initialization parameters of SSHClient, sample:
                hostname='10.121.12.58',
                port=22,
                username='root',
                password='user@root'

        @return: SSHClient(**@param(params))
        """

    @staticmethod
    def cmd(
            command: str,
            *,
            iname: 'paramiko.SSHClient' = None
    ) -> str:
        """Executes @param(command) and
        returns the result of the command."""

    @staticmethod
    def put_file(
            local: str = None,
            remote: str = None,
            *,
            iname: 'paramiko.SSHClient' = None
    ):
        """
        Sends a file to the remote host.
        @param local:  Local file path.
        @param remote: Remote file path.
        """

    @staticmethod
    def get_file(
            remote: str = None,
            local: str = None,
            *,
            iname: 'paramiko.SSHClient' = None
    ):
        """
        Pull a file from the remote host.
        @param remote: Remote file path.
        @param local:  Local file path.
        """


class itime:
    """isddc_time"""

    class Now:
        """
        Get a current time, usage:
            now_int: int = itime.Now.int
            now_str: str = itime.Now.str
        """
        str: str
        int: int
        float: float
        struct: tuple

        @staticmethod
        def strf(format: 'AnyStr') -> 'AnyStr':
            """
            @param format: Time format.
            @return:       String format time.
            """

    class Future(Now):
        """
        Get a point in time in the future, usage:
            future_time_point_1hour_int: int = itime.Future('1h').int
            future_time_point_1minute_str: str = itime.Future(60).str
        """

        def __init__(self, future_time: 'Union[int, float, str]'):
            """
            @param future_time: How long after.
            """

    class Past(Now):
        """
        Get a point in time in the past, , usage:
            past_time_point_1hour_int: int = itime.Future('1h').int
            past_time_point_1minute_str: str = itime.Future(60).str
        """

        def __init__(self, past_time: 'Union[int, float, str]'):
            """
            @param past_time: How long before.
            """

    @staticmethod
    def stamp2str(
            stamp: 'Union[int, float]' = None,
            format: str = '%F %T',
            *,
            gmt: bool = False
    ) -> str:
        """timestamp to string time"""

    @staticmethod
    def str2stamp(
            string: str,
            format: str = '%Y-%m-%d %H:%M:%S'
    ) -> float:
        """string time to timestamp"""

    @staticmethod
    def Time2Second(unit_time: str) -> 'Union[int, float]':
        """
        Convert a unit time to a second.

        @param unit_time: Can be: 1d, 1h, 1m, 1s, 1h30m, 1.5h
        @return:          Number of seconds.

        Usage:
            second_1d: int = itime.Time2Second('1d')
            second_1h30m: int = itime.Time2Second('1h30m')

        Contrary to the 'Second2Time' function below.
        """

    @staticmethod
    def Second2Time(sec: 'Union[int, float]') -> str:
        """
        Convert a second to a unit time.

        @param sec: A Number of seconds.
        @return:    A unit time of string.

        Contrary to the 'Time2Second' function below.
        """


import sys
from typing import Any, AnyStr, Union, Iterator, Callable


class ___________I___________S___________D___________D___________C___________:
    class DenyImport:
        def __init__(self, imodule: str):
            self.imodule = imodule

        def __getattribute__(self, name=None):
            if name == 'imodule':
                return super().__getattribute__(name)
            raise ModuleNotFoundError(
                f'To use this function, you need to '
                f'install the "{self.imodule}" package first.')

        __call__ = __getitem__ = __getattribute__

    def __new__(cls, iname: str = None, **params):
        iobj = cls.__init__(iname, **params)
        iname and setattr(cls, iname, iobj)
        return iobj

    for iname, icls in globals().items():
        if iname[0] == 'i':
            module_func: list = icls.__doc__.strip().split('.')
            try:
                __import__(module_func[0])
                imodule = sys.modules[module_func[0]]
            except ModuleNotFoundError:
                nonuse = DenyImport(module_func[0])
                setattr(sys.modules[__name__], iname, nonuse)
                continue
            if len(module_func) == 1:
                for attr in dir(icls):
                    if attr[0] != '_':
                        ifunc = getattr(imodule, attr, None)
                        setattr(icls, attr, ifunc)
                icls.__new__ = __new__
                icls.__init__ = imodule.__init__
            else:
                ifunc = getattr(imodule, module_func[1])
                setattr(sys.modules[__name__], iname, ifunc)


inf: idict
